# Library for communicating with an EFA Kit from Python
# Written by Kevin Ivarsen
# Copyright 2014-2018 PlaneWave Instruments

import struct
import time

import serial # Requires the pyserial library to be installed

def main():
    """
    Sample usage of the EfaSession class
    """

    SERIAL_PORT = "/dev/ttyUSB0"  # This might be a typical value in Linux
    #SERIAL_PORT = "COM8"  # This might be a typical value in Windows

    efa = EfaSession(SERIAL_PORT)

    print("Version:")
    print(efa.getVersion())

    print("Temperatures:")
    print(efa.getTemperature(TempSensor.PRIMARY))
    print(efa.getTemperature(TempSensor.AMBIENT))

    print("Current focuser position:")
    print(efa.getMotorPosition().parseData())

    print("Current rotator position:")
    print(efa.getMotorPosition(True).parseData())

    raw_input("Press enter to rack out the focuser")
    efa.trackPositive(efa.ticksPerSecondToTrackRate(50000))
    raw_input("Press enter to stop the focuser")
    efa.trackPositive(0)
    raw_input("Press enter to rack in the focuser")
    efa.trackNegative(efa.ticksPerSecondToTrackRate(50000))
    raw_input("Press enter to stop the focuser")
    efa.trackNegative(0)
    print("Focuser currently at:", efa.getMotorPosition().parseData())
    while True:
        target = raw_input("Enter a new target position:")
        try:
            target = int(target)
            break
        except:
            print("Invalid number")
            pass
    print("Focuser goto position", target)
    # NOTE: The maximum speed for the Goto command is fairly low.
    # If you need to move a long distance, it is better to use
    # trackPositive() / trackNegative() to move at higher speed,
    # monitor the current position, and when close enough,
    # send gotoPos2().
    efa.gotoPos2(target)
    while efa.isGotoOver().parseData() == 0:
        print(efa.getMotorPosition().parseData())
        time.sleep(0.5)
    print("Focuser goto finished")


    raw_input("Press enter to move rotator in positive direction")
    efa.trackPositive(efa.ticksPerSecondToTrackRate(50000), True)
    raw_input("Press enter to stop the rotator")
    efa.trackPositive(0, True)
    raw_input("Press enter to move rotator in negative direction")
    efa.trackNegative(efa.ticksPerSecondToTrackRate(50000), True)
    raw_input("Press enter to stop the rotator")
    efa.trackNegative(0, True)
    print("Rotator currently at:", efa.getMotorPosition(True).parseData())
    while True:
        target = raw_input("Enter a new target position:")
        try:
            target = int(target)
            break
        except:
            print("Invalid number")
            pass
    print("Rotator goto position", target)
    # See earlier note about how to do a faster Goto
    efa.gotoPos2(target, True)
    while efa.isGotoOver(True).parseData() == 0:
        print(efa.getMotorPosition(True).parseData())
        time.sleep(0.5)
    print("Rotator goto finished")






class Enum:
    """
    Utility class for creating a set of named constants.
    Name can be converted to value, and value and be converted to string name.

    Example:

    commands = Constants(
        GET_TEMP = 0x01,
        GET_RATE = 0x02,
        SET_RATE = 0x03
        )

    sendCommand(commands.GET_TEMP)
    print(commands.getName(0x01)) # GET_TEMP
    """

    def __init__(self, **kwargs):
        self._namesToValues = kwargs
        self._valuesToNames = dict()
        for k,v in self._namesToValues.items():
            self._valuesToNames[v] = k

    def getName(self, value):
        return self._valuesToNames.get(value, "UNKNOWN(%r)" % value)
    
    def __getattr__(self, name):
        return self._namesToValues[name]


# Device addresses recognized by the EFA
Address = Enum(
    PC = 0x20,       # User's computer
    HC = 0x0D,       # EFA hand controller
    FOC_TEMP = 0x12, # EFA focus motor and temp sensors
    ROT_FAN = 0x13,  # EFA rotator motor and fan control
    DELTA_T = 0x32   # Delta-T dew heater
)

# Temperature sensors.
# NOTE: Not all telescope configurations include all temp sensors!
TempSensor = Enum(
    PRIMARY = 0,
    AMBIENT = 1,
    SECONDARY = 2,
    BACKPLATE = 3,
    M3 = 4
)

# Commands recognized by the EFA. See protocol documentation for more details.
Command = Enum(
    MTR_GET_POS = 0x01,
    MTR_GOTO_POS2 = 0x17,
    MTR_OFFSET_CNT = 0x04,
    MTR_GOTO_OVER = 0x13,
    MTR_PTRACK = 0x06,
    MTR_NTRACK = 0x07,
    MTR_SLEWLIMITMIN = 0x1A,
    MTR_SLEWLIMITMAX = 0x1B,
    MTR_SLEWLIMITGETMIN = 0x1C,
    MTR_SLEWLIMITGETMAX = 0x1D,
    MTR_PMSLEW_RATE = 0x24,
    MTR_NMSLEW_RATE = 0x25,
    TEMP_GET = 0x26,
    FANS_SET = 0x27,
    FANS_GET = 0x28,
    MTR_GET_CALIBRATION_STATE = 0x30,
    MTR_SET_CALIBRATION_STATE = 0x31,
    MTR_GET_STOP_DETECT = 0xEE,
    MTR_STOP_DETECT = 0xEF,
    MTR_GET_APPROACH_DIRECTION = 0xFC,
    MTR_APPROACH_DIRECTION = 0xFD,
    GET_VERSION = 0xFE,

)

class IRF90:
    ROTATOR_TICKS_PER_DEG = 36408.0
    FOCUSER_TICKS_PER_MICRON = 207.288

    def __init__(self, efaSession):
        self.efa = efaSession

        self.goto2ThresholdTicks = 50000
        self.gotoTicksPerSec = 50000

    def getRotatorPosDegs(self):
        posTicks = self.efa.getMotorPosition(True).parseData()
        posDegs = posTicks / self.ROTATOR_TICKS_PER_DEG
        return posDegs

    def getFocuserPosMicrons(self):
        posTicks = self.efa.getMotorPosition(False).parseData()
        posMicrons = posTicks / self.FOCUSER_TICKS_PER_MICRON
        return posMicrons

    def focuserGoto(self, targetMicrons):
        targetTicks = int(targetMicrons * self.FOCUSER_TICKS_PER_MICRON)
        self.axisGoto(targetTicks, False)

    def rotatorGoto(self, targetDegs):
        targetTicks = int(targetDegs * self.ROTATOR_TICKS_PER_DEG)
        self.axisGoto(targetTicks, True)

    def rotatorSetRate(self, rateDegsPerSec):
        rateTicksPerSec = int(rateDegsPerSec * self.ROTATOR_TICKS_PER_DEG)
        self.efa.trackTicksPerSec(rateTicksPerSec, True)

    def focuserFindHome(self):
        self.findHome(False)

    def rotatorFindHome(self):
        self.findHome(True)

    def axisGoto(self, targetTicks, useRotator):
        currentPosTicks = self.efa.getMotorPosition(useRotator).parseData()

        if targetTicks > currentPosTicks:
            slewDirection = 1
        else:
            slewDirection = -1

        targetBeforeGoto2 = targetTicks - slewDirection*self.goto2ThresholdTicks

        remainingTicksBeforeStop = slewDirection * (targetBeforeGoto2 - currentPosTicks)

        if remainingTicksBeforeStop > 0:
            ticksPerSec = slewDirection * self.gotoTicksPerSec
            self.efa.trackTicksPerSec(ticksPerSec, useRotator)

            while True:
                currentPosTicks = self.efa.getMotorPosition(useRotator).parseData()
                remainingTicksBeforeStop = slewDirection * (targetBeforeGoto2 - currentPosTicks)

                if useRotator:
                    print("%.4f" % (currentPosTicks/self.ROTATOR_TICKS_PER_DEG))
                else:
                    print("%.4f" % (currentPosTicks/self.FOCUSER_TICKS_PER_MICRON))

                isGotoOver = self.efa.isGotoOver(useRotator).parseData()
                # For a velocity command, 0xFF = moving, 0xFE = stopped/aborted
                if isGotoOver == 0xFE:
                    print("ABORTED")
                    return

                if remainingTicksBeforeStop < 0:
                    self.efa.stop(useRotator)
                    break

                time.sleep(0.1)

        self.efa.gotoPos2(targetTicks, useRotator)

        while True:
            if useRotator:
                print(self.getRotatorPosDegs())
            else:
                print(self.getFocuserPosMicrons())

            isGotoOver = self.efa.isGotoOver(useRotator).parseData()
            # For a Goto command, 0x00 = Goto in progress, 0xFF = Goto finished, 0xFE = aborted
            if isGotoOver == 0xFF:
                print("Finished!")
                return
            elif isGotoOver == 0xFE:
                print("ABORTED")
                return

            time.sleep(0.5)

    def findHome(self, useRotator):
        print("Finding hardstop in negative direction...")
        self.efa.trackTicksPerSec(-50000, useRotator)

        time.sleep(0.2)
        while True:
            pos = self.efa.getMotorPosition(useRotator).parseData()
            print(pos)

            isGotoOver = self.efa.isGotoOver(useRotator).parseData()

            print("isGotoOver", isGotoOver)
            if isGotoOver == 0xFE:
                print("Found hardstop")
                break

            time.sleep(0.2)

        print("Resetting encoder from %f to 0" % pos)
        self.efa.setEncoder(0, useRotator)

        print("Moving away from hardstop")
        self.efa.gotoPos2(50000, useRotator)

        while True:
            pos = self.efa.getMotorPosition(useRotator).parseData()
            print(pos)

            isGotoOver = self.efa.isGotoOver(useRotator).parseData()
            if isGotoOver != 0:
                print("Done")
                break








class EfaSession:
    BAUD_RATE = 19200
    SOM = 0x3B # Start-of-message byte

    def __init__(self, comPort, isDeltaT=False):
        self.port = serial.Serial(comPort, self.BAUD_RATE)

        # Delta-T differs slightly from the EFA.
        # For example, the Delta-T does not use the RTS/CTS line to take
        # control of the serial bus, and the Delta-T does not echo back
        # everything received from the PC.
        self.isDeltaT = isDeltaT

        self.debug = True # False # If true, print(some extra logging messages)

        self.abortOnTimeout = True

        self.echoReceivedBytes = False # If true, echo each incoming byte back to serial port
                                       # (Used to simulate an EFA)

        if not isDeltaT:
            # Lower the RTS line
            # Otherwise we monopoloze the serial network and handcontroller 
            # doesn't work
            self.port.setRTS(False)

        if "setTimeout" in dir(self.port):
            # Earlier versions of pyserial used setTimeout()
            self.port.setTimeout(1)
        else:
            # More recent versions use timeout
            self.port.timeout = 1

    def close(self):
        self.port.close()

    ### High-level EFA commands ###

    def getVersion(self):
        response = self.sendReceive(Address.FOC_TEMP,
                                    Command.GET_VERSION)
        version = "%d.%d" % (response.data[0], response.data[1])
        return version

    def gotoPos2(self, target, useRotator=False):
        if useRotator:
            address = Address.ROT_FAN
        else:
            address = Address.FOC_TEMP

        response = self.sendReceive(address,
                                    Command.MTR_GOTO_POS2,
                                    *EfaPacket.intTo3Bytes(target))

        return response

    def isGotoOver(self, useRotator=False):
        if useRotator:
            address = Address.ROT_FAN
        else:
            address = Address.FOC_TEMP

        response = self.sendReceive(address,
                                    Command.MTR_GOTO_OVER)

        return response

    def slewPositive(self, rate, useRotator=False):
        """
        rate = 0 (stopped) to 9 (fastest)
        useRotator: if true, move Rotator motor; else move Focus motor

        rates in motor ticks/sec, measured on an IRF90:
            9	69907
            8	46603
            7	23306
            6	12459
            5	3117
            4	1557
            3	779
            2	195
            1	96
            0	0
        """

        if useRotator:
            address = Address.ROT_FAN
        else:
            address = Address.FOC_TEMP

        response = self.sendReceive(address,
                                    Command.MTR_PMSLEW_RATE,
                                    rate)

        return response

    def slewNegative(self, rate, useRotator=False):
        """
        rate = 0 (stopped) to 9 (fastest)
        useRotator: if true, move Rotator motor; else move Focus motor
        """

        if useRotator:
            address = Address.ROT_FAN
        else:
            address = Address.FOC_TEMP

        response = self.sendReceive(address,
                                    Command.MTR_NMSLEW_RATE,
                                    rate)

        return response

    def trackPositive(self, rate, useRotator=False):
        if useRotator:
            address = Address.ROT_FAN
        else:
            address = Address.FOC_TEMP

        byte1, byte2, byte3 = EfaPacket.intTo3Bytes(rate)
        response = self.sendReceive(address,
                                    Command.MTR_PTRACK,
                                    byte1, byte2, byte3)

        return response

    def trackNegative(self, rate, useRotator=False):
        if useRotator:
            address = Address.ROT_FAN
        else:
            address = Address.FOC_TEMP

        byte1, byte2, byte3 = EfaPacket.intTo3Bytes(rate)
        response = self.sendReceive(address,
                                    Command.MTR_NTRACK,
                                    byte1, byte2, byte3)

        return response

    def track(self, rate, useRotator=False):
        if rate < 0:
            self.trackNegative(abs(rate), useRotator)
        else:
            self.trackPositive(rate, useRotator)

    def trackNegativeTicksPerSec(self, ticksPerSec, useRotator=False):
        rate = int(ticksPerSecondToTrackRate(ticksPerSec))
        self.trackNegative(rate, useRotator)

    def trackPositiveTicksPerSec(self, ticksPerSec, useRotator=False):
        rate = int(ticksPerSecondToTrackRate(ticksPerSec))
        self.trackPositive(rate, useRotator)

    def trackTicksPerSec(self, ticksPerSec, useRotator=False):
        rate = int(ticksPerSecondToTrackRate(ticksPerSec))
        self.track(rate, useRotator)


    def stop(self, useRotator=False):
        self.trackPositive(0, useRotator)

    def ticksPerSecondToTrackRate(self, ticksPerSec):
        # Given a desired number of encoder ticks per second,
        # return the native rate value to send to trackPositive()
        # or trackNegative()
        return int(ticksPerSec * 79.101)


    def getMotorPosition(self, useRotator=False):
        if useRotator:
            address = Address.ROT_FAN
        else:
            address = Address.FOC_TEMP

        response = self.sendReceive(address,
                                    Command.MTR_GET_POS)
        return response

    def setEncoder(self, newEncoderValueTicks, useRotator=False):
        if useRotator:
            address = Address.ROT_FAN
        else:
            address = Address.FOC_TEMP

        byte1, byte2, byte3 = EfaPacket.intTo3Bytes(newEncoderValueTicks)

        response = self.sendReceive(address, 
                                    Command.MTR_OFFSET_CNT, byte1, byte2, byte3)

        return response

    def getFanState(self):
        response = self.sendReceive(Address.ROT_FAN,
                                    Command.FANS_GET)

        return response

    def setFanState(self, isOn):
        if isOn:
            value = 1
        else:
            value = 0

        response = self.sendReceive(Address.ROT_FAN,
                                    Command.FANS_SET,
                                    value)

        return response

    def getTemperature(self, tempAddress):
        """
        tempAddress:
          0 = Primary
          1 = Ambient
          2 = Secondary
          3 = Backplate
          4 = M3
        """

        response = self.sendReceive(Address.FOC_TEMP,
                                    Command.TEMP_GET,
                                    tempAddress)

        if self.debug:
            print(response)

        if tuple(response.data) == (127, 127):
            return None # no temp sensor at this address
        else:
            # Temperature is returned LSB-first
            return (256*response.data[1] + response.data[0]) / 16.0



    ### Low-level functions ###

    def send(self, receiverAddress, command, *data):
        packet = EfaPacket(Address.PC, receiverAddress, command, *data)

        if self.debug:
            print("Send: %r" % packet.toHexString())

        if not self.isDeltaT:
            while self.port.getCTS() == True:
                #print("Waiting for CTS...")
                pass

        if not self.isDeltaT:
            self.port.setRTS(True)

        self.port.write(packet.toBytes())

    def sendReceive(self, receiverAddress, command, *data):
        """
        Send a packet, and return the response packet
        """

        self.send(receiverAddress, command, *data)

        if not self.isDeltaT:
            # An EFA kit echoes everything back, so expect to
            # read the packet that we just sent. A Delta T on the
            # other hand, will simply send back the response straight away
            ackPacket = self.readNextPacket()
            if ackPacket is None:
                print("(timeout on ack)")
                return None # Timeout

            if self.debug:
                print("Echo: %r" % ackPacket.toHexString())

        if not self.isDeltaT:
            self.port.setRTS(False)

        while True:
            response = self.readNextPacket()
            if response is None:
                print("(timeout on response packet)")
                return None # Timeout

            if self.debug:
                print("Resp: %r" % response.toHexString())

            if self.debug:
                print("Response: " + response.description())

            if response.receiverAddress == Address.PC:
                return response
            elif self.debug:
                print("  (Ignored by PC)")


    def readNextByte(self):
        c = self.port.read(1)
        if c == "":
            return None
        if self.echoReceivedBytes:
            self.port.write(c)
        return ord(c)

    def readNextPacket(self):
        # scan for SOM
        while True:
            c = self.readNextByte()
            if c == None:
                if self.abortOnTimeout:
                    return None # No packet on timeout
                else:
                    print("...")
                    continue

            if c == EfaSession.SOM:
                break

            if self.debug:
                print("Ignore: %02X (%s)" % (c, chr(c)))

        numBytes = self.readNextByte()
        sourceAddress = self.readNextByte()
        receiverAddress = self.readNextByte()
        command = self.readNextByte()
        
        data = []
        for i in range(numBytes-3): # After SrcAdr, DstAdr, and Cmd, the rest of the numbered bytes are data
            data.append(self.readNextByte())

        checksum = self.readNextByte()

        packet = EfaPacket(sourceAddress, receiverAddress, command, *data)
        packet.receivedChecksumByte = checksum

        return packet



class EfaPacket:
    def __init__(self, sourceAddress, receiverAddress, command, *dataBytes):
        self.sourceAddress = sourceAddress
        self.receiverAddress = receiverAddress
        self.command = command
        self.data = dataBytes

        self.receivedChecksumByte = None # Caller can set this when a received packet includes a checksum

        # Make sure things are in range
        for byte in self.data:
            if byte < 0 or byte > 255:
                raise Exception("EfaPacket data byte out of range: cmd=%02X, data=%r" % (self.command, self.data))

    def calculatedChecksum(self):
        return ord(self.toBytes()[-1])

    def isChecksumOk(self):
        """
        Depends on caller first setting receivedChecksumByte
        """

        print("CHK Calculated: %02X  Received: %02X" % (self.calculatedChecksum(), self.receivedChecksumByte))
        return self.receivedChecksumByte == self.calculatedChecksum()

    def toBytes(self):
        numBytes = 3 + len(self.data)

        packetFormat = "BBBBB" + ("B" * len(self.data)) + "B"

        packetFields = [EfaSession.SOM, numBytes, self.sourceAddress, self.receiverAddress, self.command]
        packetFields.extend(self.data)

        checksum = sum(packetFields[1:]) # Sum of fields, excluding SOM (and CHK, which we haven't included yet of course)
        checksum = -checksum & 0xFF  # LSB of the two's complement
        packetFields += [checksum]

        return struct.pack(packetFormat, *packetFields)

    def toHexString(self):
        bytes = self.toBytes()
        hexString = " ".join(["%02X" % ord(x) for x in bytes])
        return hexString

    def dataAsByteString(self):
        """
        Return the data array as a string of bytes
        """

        return "".join([chr(x) for x in self.data])

    def parseData(self):
        """
        Assuming the data payload contains an integer encoded
        as one or more MSB-first bytes, return the value
        """

        value = 0

        for byte in self.data:
            value = value << 8 | byte

        return value

    def description(self):
        if self.receivedChecksumByte is None:
            checksumOk = "??"
        elif self.isChecksumOk():
            checksumOk = "OK"
        else:
            checksumOk = "FAILED"

        if len(self.data) <= 3:
            parsedValue = self.parseData()
        else:
            parsedValue = "N/A"

        return "%s -> %s: %s %r  [value: %s; checksum: %s]\n[hex: %s]" % (
                Address.getName(self.sourceAddress), 
                Address.getName(self.receiverAddress), 
                Command.getName(self.command),
                self.data,
                parsedValue,
                checksumOk,
                self.toHexString())

    def __str__(self):
        return self.description()

    def __repr__(self):
        return str(self)

    @staticmethod
    def intTo3Bytes(value):
        """
        Values sent to the EFA are commonly encoded as 3 bytes
        packed MSB first. Given an integer value, return a
        tuple of 3 bytes that encode that value
        """

        lsb = value & 0xFF
        value >>= 8
        middle = value & 0xFF
        value >>= 8
        msb = value & 0xFF
        return (msb, middle, lsb)

def rawTemperatureToCelsius(rawTemp, byte2=None):
    """
    Accepts as arguments either (MSB, LSB) or a single raw temperature value
    """

    # Combine bytes if two args provided
    if byte2 is not None:
        rawTemp = rawTemp + byte2*256

    if rawTemp & 0x8000: # Is high bit set? If so, value is negative
        rawTemp = rawTemp - 0x10000

    # Raw temperature is expressed in 16ths of a degree C
    # A simple division gives us C
    return rawTemp / 16.0

def celsiusToRawTemperature(celsius):
    rawTemp = int(celsius*16)
    if rawTemp < 0:
        rawTemp += 65536
    return rawTemp

def celsiusToRawTemperatureBytes(celsius):
    rawTemp = celsiusToRawTemperature(celsius)
    byte1, byte2, byte3 = EfaPacket.intTo3Bytes(rawTemp)
    return byte3, byte2 # LSB, MSB (which seems to be the format produced by EFA hardware

def trackRateToTicksPerSecond(trackRate):
    """
    Convert the value given to MTR_PTRACK / MTR_NTRACK to
    some number of ticks per second. Determined empirically on 
    the prototype IRF90
    """

    return trackRate / 79.101

def ticksPerSecondToTrackRate(ticksPerSecond):
    """
    Find the value to be given to MTR_PTRACK / MTR_NTRACK to
    achieve some number of ticks per second. Determined empirically on 
    the prototype IRF90
    """
    return ticksPerSecond * 79.101

if __name__ == "__main__":
    main()
